//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/FormFactorItemCatalog.cpp
//! @brief     Implements class FormFactorItemCatalog
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/FormFactorItemCatalog.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Sample/FormFactorItems.h"

FormFactorItem* FormFactorItemCatalog::create(Type type)
{
#define CREATE(t)                                                                                  \
    case Type::t:                                                                                  \
        return new t##Item()

    switch (type) {
        CREATE(AnisoPyramid);
        CREATE(BarGauss);
        CREATE(BarLorentz);
        CREATE(Box);
        CREATE(Cone);
        CREATE(Cone6);
        CREATE(Cuboctahedron);
        CREATE(Cylinder);
        CREATE(Dodecahedron);
        CREATE(EllipsoidalCylinder);
        CREATE(FullSphere);
        CREATE(FullSpheroid);
        CREATE(HemiEllipsoid);
        CREATE(Icosahedron);
        CREATE(Prism3);
        CREATE(Prism6);
        CREATE(Pyramid);
        CREATE(CosineRippleBox);
        CREATE(CosineRippleGauss);
        CREATE(CosineRippleLorentz);
        CREATE(SawtoothRippleBox);
        CREATE(SawtoothRippleGauss);
        CREATE(SawtoothRippleLorentz);
        CREATE(Tetrahedron);
        CREATE(TruncatedCube);
        CREATE(TruncatedSphere);
        CREATE(TruncatedSpheroid);
    default:
        ASSERT(false);
    }

#undef CREATE
}

QVector<FormFactorItemCatalog::Type> FormFactorItemCatalog::types()
{
    return {Type::AnisoPyramid,
            Type::BarGauss,
            Type::BarLorentz,
            Type::Box,
            Type::Cone,
            Type::Cone6,
            Type::Cuboctahedron,
            Type::Cylinder,
            Type::Dodecahedron,
            Type::EllipsoidalCylinder,
            Type::FullSphere,
            Type::FullSpheroid,
            Type::HemiEllipsoid,
            Type::Icosahedron,
            Type::Prism3,
            Type::Prism6,
            Type::Pyramid,
            Type::CosineRippleBox,
            Type::CosineRippleGauss,
            Type::CosineRippleLorentz,
            Type::SawtoothRippleBox,
            Type::SawtoothRippleGauss,
            Type::SawtoothRippleLorentz,
            Type::Tetrahedron,
            Type::TruncatedCube,
            Type::TruncatedSphere,
            Type::TruncatedSpheroid};
}

QVector<FormFactorItemCatalog::Type> FormFactorItemCatalog::hardParticleTypes()
{
    return {Type::AnisoPyramid,
            Type::Box,
            Type::Cone,
            Type::Cone6,
            Type::Cuboctahedron,
            Type::Cylinder,
            Type::Dodecahedron,
            Type::EllipsoidalCylinder,
            Type::FullSphere,
            Type::FullSpheroid,
            Type::HemiEllipsoid,
            Type::Icosahedron,
            Type::Prism3,
            Type::Prism6,
            Type::Pyramid,
            Type::Tetrahedron,
            Type::TruncatedCube,
            Type::TruncatedSphere,
            Type::TruncatedSpheroid};
}


QVector<FormFactorItemCatalog::Type> FormFactorItemCatalog::rippleTypes()
{
    return {Type::Box,
            Type::BarGauss,
            Type::BarLorentz,
            Type::CosineRippleBox,
            Type::CosineRippleGauss,
            Type::CosineRippleLorentz,
            Type::SawtoothRippleBox,
            Type::SawtoothRippleGauss,
            Type::SawtoothRippleLorentz};
}

FormFactorItemCatalog::UiInfo FormFactorItemCatalog::uiInfo(Type type)
{
    auto createUiInfo = [](const QString& menuEntry, const QString& iconPath,
                           const QString& description) {
        UiInfo info;
        info.menuEntry = menuEntry;
        info.iconPath = iconPath;
        info.description = description;
        return info;
    };

    switch (type) {
    case Type::AnisoPyramid:
        return createUiInfo("Anisotropic pyramid",
                            ":/SampleDesignerToolbox/images/ff_AnisoPyramid_64x64.png",
                            "Truncated pyramid with a rectangular base");

    case Type::BarGauss:
        return createUiInfo("BarGauss", ":/SampleDesignerToolbox/images/ff_Box_64x64.png",
                            "Rectangular cuboid");

    case Type::BarLorentz:
        return createUiInfo("BarLorentz", ":/SampleDesignerToolbox/images/ff_Box_64x64.png",
                            "Rectangular cuboid");

    case Type::Box:
        return createUiInfo("Box", ":/SampleDesignerToolbox/images/ff_Box_64x64.png",
                            "Rectangular cuboid");

    case Type::Cone:
        return createUiInfo("Cone", ":/SampleDesignerToolbox/images/ff_Cone_64x64.png",
                            "Truncated cone with circular base");

    case Type::Cone6:
        return createUiInfo("Cone6", ":/SampleDesignerToolbox/images/ff_Cone6_64x64.png",
                            "A truncated pyramid, based on a regular hexagon");

    case Type::Cuboctahedron:
        return createUiInfo("Cuboctahedron",
                            ":/SampleDesignerToolbox/images/ff_Cuboctahedron_64x64.png",
                            "Compound of two truncated pyramids with a common square base "
                            "and opposite orientations");

    case Type::Dodecahedron:
        return createUiInfo("Dodecahedron",
                            ":/SampleDesignerToolbox/images/ff_Dodecahedron_64x64.png",
                            "Dodecahedron");

    case Type::Cylinder:
        return createUiInfo("Cylinder", ":/SampleDesignerToolbox/images/ff_Cylinder_64x64.png",
                            "Cylinder with a circular base");

    case Type::EllipsoidalCylinder:
        return createUiInfo("Ellipsoidal cylinder",
                            ":/SampleDesignerToolbox/images/ff_EllipsoidalCylinder_64x64.png",
                            "Cylinder with an ellipse cross section");

    case Type::FullSphere:
        return createUiInfo("Full sphere", ":/SampleDesignerToolbox/images/ff_FullSphere_64x64.png",
                            "Full sphere");

    case Type::FullSpheroid:
        return createUiInfo(
            "Full spheroid", ":/SampleDesignerToolbox/images/ff_FullSpheroid_64x64.png",
            "Full spheroid, generated by rotating an ellipse around the vertical axis");

    case Type::HemiEllipsoid:
        return createUiInfo("Hemi ellipsoid",
                            ":/SampleDesignerToolbox/images/ff_HemiEllipsoid_64x64.png",
                            "A horizontally oriented ellipsoid, truncated at the central plane");

    case Type::Icosahedron:
        return createUiInfo("Icosahedron",
                            ":/SampleDesignerToolbox/images/ff_Icosahedron_64x64.png",
                            "Icosahedron");

    case Type::Prism3:
        return createUiInfo("Prism3", ":/SampleDesignerToolbox/images/ff_Prism3_64x64.png",
                            "Prism with an equilateral triangle base");

    case Type::Prism6:
        return createUiInfo("Prism6", ":/SampleDesignerToolbox/images/ff_Prism6_64x64.png",
                            "Prism with a regular hexagonal base");

    case Type::Pyramid:
        return createUiInfo("Pyramid", ":/SampleDesignerToolbox/images/ff_Pyramid_64x64.png",
                            "Truncated pyramid with a square base");

    case Type::CosineRippleBox:
        return createUiInfo("CosineRippleBox",
                            ":/SampleDesignerToolbox/images/ff_Ripple1_64x64.png",
                            "Particle with a cosine profile and a rectangular base");

    case Type::CosineRippleGauss:
        return createUiInfo("CosineRippleGauss",
                            ":/SampleDesignerToolbox/images/ff_Ripple1_64x64.png",
                            "Particle with a cosine profile and a rectangular base");

    case Type::CosineRippleLorentz:
        return createUiInfo("CosineRippleLorentz",
                            ":/SampleDesignerToolbox/images/ff_Ripple1_64x64.png",
                            "Particle with a cosine profile and a rectangular base");

    case Type::SawtoothRippleBox:
        return createUiInfo("SawtoothRippleBox",
                            ":/SampleDesignerToolbox/images/ff_Ripple2_64x64.png",
                            "Particle with an asymmetric triangle profile and a rectangular base");

    case Type::SawtoothRippleGauss:
        return createUiInfo("SawtoothRippleGauss",
                            ":/SampleDesignerToolbox/images/ff_Ripple2_64x64.png",
                            "Particle with an asymmetric triangle profile and a rectangular base");

    case Type::SawtoothRippleLorentz:
        return createUiInfo("SawtoothRippleLorentz",
                            ":/SampleDesignerToolbox/images/ff_Ripple2_64x64.png",
                            "Particle with an asymmetric triangle profile and a rectangular base");

    case Type::Tetrahedron:
        return createUiInfo(
            "Tetrahedron", ":/SampleDesignerToolbox/images/ff_Tetrahedron_64x64.png",
            "Truncated polyhedron with equilateral triangle base and cropped side faces");

    case Type::TruncatedCube:
        return createUiInfo("Truncated cube",
                            ":/SampleDesignerToolbox/images/ff_TruncatedCube_64x64.png",
                            "A cube whose eight vertices have been removed");

    case Type::TruncatedSphere:
        return createUiInfo("Truncated sphere",
                            ":/SampleDesignerToolbox/images/ff_TruncatedSphere_64x64.png",
                            "Spherical dome");

    case Type::TruncatedSpheroid:
        return createUiInfo("Truncated spheroid",
                            ":/SampleDesignerToolbox/images/ff_TruncatedSpheroid_64x64.png",
                            "Spheroidal dome");

    default:
        ASSERT(false);
    }
}

FormFactorItemCatalog::Type FormFactorItemCatalog::type(const FormFactorItem* item)
{
    ASSERT(item);

#define CHECK(type)                                                                                \
    if (dynamic_cast<const type##Item*>(item))                                                     \
    return Type::type

    CHECK(AnisoPyramid);
    CHECK(BarGauss);
    CHECK(BarLorentz);
    CHECK(Box);
    CHECK(Cone);
    CHECK(Cone6);
    CHECK(Cuboctahedron);
    CHECK(Cylinder);
    CHECK(Dodecahedron);
    CHECK(EllipsoidalCylinder);
    CHECK(FullSphere);
    CHECK(FullSpheroid);
    CHECK(HemiEllipsoid);
    CHECK(Icosahedron);
    CHECK(Prism3);
    CHECK(Prism6);
    CHECK(Pyramid);
    CHECK(CosineRippleBox);
    CHECK(CosineRippleGauss);
    CHECK(CosineRippleLorentz);
    CHECK(SawtoothRippleBox);
    CHECK(SawtoothRippleGauss);
    CHECK(SawtoothRippleLorentz);
    CHECK(Tetrahedron);
    CHECK(TruncatedCube);
    CHECK(TruncatedSphere);
    CHECK(TruncatedSpheroid);
#undef CHECK
    ASSERT(false);
}

QString FormFactorItemCatalog::menuEntry(const FormFactorItem* item)
{
    return uiInfo(type(item)).menuEntry;
}
