//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/ItemWithParticles.cpp
//! @brief     Implements class ItemWithParticles
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/ItemWithParticles.h"
#include "Base/Vector/RotMatrix.h"
#include "GUI/Model/Sample/RotationItemCatalog.h"
#include "GUI/Model/Session/Serializer.h"
#include "Sample/Particle/IParticle.h"
#include "Sample/Scattering/Rotations.h"


ItemWithParticles::ItemWithParticles(const QString& abundanceTooltip,
                                     const QString& positionTooltip)
{
    m_abundance.init("Abundance", abundanceTooltip, 1.0, Unit::unitless, 3,
                     RealLimits::limited(0.0, 1.0), "abundance");
    m_position.init("Position Offset", positionTooltip, Unit::nanometer, "pos");
    m_rotation.init<RotationItemCatalog>("Rotation", "", "rotation");
}

DoubleDescriptor ItemWithParticles::abundance() const
{
    return m_abundance;
}

void ItemWithParticles::setAbundance(const double abundance)
{
    m_abundance.set(abundance);
}

R3 ItemWithParticles::position() const
{
    return m_position;
}

void ItemWithParticles::setPosition(const R3& position)
{
    m_position.set(position);
}

VectorDescriptor ItemWithParticles::positionVector() const
{
    return m_position;
}

SelectionDescriptor<RotationItem*> ItemWithParticles::rotationMethod()
{
    return m_rotation;
}

void ItemWithParticles::setRotation(RotationItem* p)
{
    m_rotation.set(p);
}

void ItemWithParticles::setTransformationInfo(IParticle* result) const
{
    result->setParticlePosition(position());
    const RotMatrix r = m_rotation.get() ? m_rotation.get()->rotation() : RotMatrix();
    if (!r.isIdentity()) {
        std::unique_ptr<IRotation> rotation(IRotation::createRotation(r));
        result->setRotation(*rotation);
    }
}
