//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/ItemWithParticlesCatalog.cpp
//! @brief     Implements class ItemWithParticlesCatalog
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/ItemWithParticlesCatalog.h"
#include "GUI/Model/Sample/MesoCrystalItem.h"
#include "GUI/Model/Sample/ParticleCompositionItem.h"
#include "GUI/Model/Sample/ParticleCoreShellItem.h"
#include "GUI/Model/Sample/ParticleItem.h"

ItemWithParticles* ItemWithParticlesCatalog::create(Type type)
{
    switch (type) {
    case Type::Particle:
        return new ParticleItem();
    case Type::Composition:
        return new ParticleCompositionItem();
    case Type::CoreShell:
        return new ParticleCoreShellItem();
    case Type::MesoCrystal:
        return new MesoCrystalItem();
    default:
        ASSERT(false);
    }
}

QVector<ItemWithParticlesCatalog::Type> ItemWithParticlesCatalog::types()
{
    return {Type::Particle, Type::Composition, Type::CoreShell, Type::MesoCrystal};
}

ItemWithParticlesCatalog::UiInfo ItemWithParticlesCatalog::uiInfo(Type type)
{
    auto createUiInfo = [](const QString& menuEntry, const QString& iconName,
                           const QString& description) {
        UiInfo info;
        info.menuEntry = menuEntry;
        info.iconPath = ":/SampleDesignerToolbox/images/" + iconName;
        info.description = description;
        return info;
    };

    switch (type) {
    case Type::Particle:
        return createUiInfo("Particle", "", ""); // particle is not on UI, only its form factor
    case Type::Composition:
        return createUiInfo("Particle Composition", "ParticleComposition_64x64.png",
                            "Composition of particles with fixed positions");
    case Type::CoreShell:
        return createUiInfo("Core shell particle", "ParticleCoreShell_64x64.png",
                            "A particle with a core/shell geometry");
    case Type::MesoCrystal:
        return createUiInfo("Meso Crystal", "Mesocrystal_64x64.png",
                            "A 3D crystal structure of nanoparticles");
    default:
        ASSERT(false);
    }
}

ItemWithParticlesCatalog::Type ItemWithParticlesCatalog::type(const ItemWithParticles* item)
{
    ASSERT(item);

    if (dynamic_cast<const ParticleItem*>(item))
        return Type::Particle;
    if (dynamic_cast<const ParticleCompositionItem*>(item))
        return Type::Composition;
    if (dynamic_cast<const MesoCrystalItem*>(item))
        return Type::MesoCrystal;
    if (dynamic_cast<const ParticleCoreShellItem*>(item))
        return Type::CoreShell;

    ASSERT(false);
}
