//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Session/Serializer.h
//! @brief     Defines class Serializer
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_SESSION_SERIALIZER_H
#define BORNAGAIN_GUI_MODEL_SESSION_SERIALIZER_H

#include "GUI/Model/Types/DoubleDescriptor.h"
#include "GUI/Model/Types/SelectionProperty.h"
#include <QXmlStreamWriter> // used in every including file

class DoubleProperty;
class VectorProperty;
class UIntProperty;

class Serializer {
public:
    Serializer(QXmlStreamWriter* w) : m_w(w) {}
    Serializer(QXmlStreamReader* r) : m_r(r) {}

    //! No meaning for writing
    void assertVersion(unsigned expectedVersion) const;
    bool isReading() const;

    void rw(const QString& tag, QString& val);
    void rw(const QString& tag, QColor& col);
    void rw(const QString& tag, bool& val);
    void rw(DoubleProperty& d);
    void rw(VectorProperty& d);
    void rw(UIntProperty& d);


    //! serialize a list with a known and fixed type which is not a base class, but the real class
    template <typename T> void rw(const QString& tag, QVector<T>& vec)
    {
        if (m_w) {
            m_w->writeStartElement(tag);
            m_w->writeAttribute("n", QString::number(vec.size()));
            int i = 0;
            for (auto* p : vec) {
                m_w->writeStartElement("E" + QString::number(i++));
                p->serialize(*this);
                m_w->writeEndElement();
            }
            m_w->writeEndElement();
        } else {
            m_r->readNextStartElement();
            assertCurrentTag(tag);
            const int n = m_r->attributes().value("n").toInt();
            for (int i = 0; i < n; i++) {
                goToStartElementOfTag("E" + QString::number(i));
                vec << new typename std::remove_pointer<T>::type();
                vec.last()->serialize(*this);
                goToEndElementOfTag("E" + QString::number(i));
            }
            goToEndElementOfTag(tag);
        }
    }

    //! serialize a list with a base class type, which can contain derived types. For this we need
    //! the catalog of possible classes/types to identify the current type and to re-create this
    //! type after serialization
    template <typename Catalog>
    void rw(const QString& tag, QVector<typename Catalog::CatalogedType*>& vec)
    {
        if (m_w) {
            m_w->writeStartElement(tag);
            m_w->writeAttribute("n", QString::number(vec.size()));
            int i = 0;
            for (auto* p : vec) {
                write<Catalog>("E" + QString::number(i++), p);
            }
            m_w->writeEndElement();
        } else {
            m_r->readNextStartElement();
            assertCurrentTag(tag);
            const int n = m_r->attributes().value("n").toInt();
            for (int i = 0; i < n; i++) {
                typename Catalog::CatalogedType* p = nullptr;
                read<Catalog>("E" + QString::number(i), p);
                vec << p;
            }
            goToEndElementOfTag(tag);
        }
    }

    template <typename Catalog> void rw(SelectionProperty<typename Catalog::CatalogedType*>& d)
    {
        if (m_w) {
            typename Catalog::CatalogedType* p = d.get();
            write<Catalog>(d.persistentTag(), p);
        } else {
            typename Catalog::CatalogedType* p = nullptr;
            read<Catalog>(d.persistentTag(), p);
            d.set(p);
        }
    }

    //! serialize item which has only one exact class, and can be present or not
    template <typename ItemClass> void rw(const QString& tag, ItemClass*& p)
    {
        if (m_w) {
            m_w->writeStartElement(tag);
            m_w->writeAttribute("valid", p ? "1" : "0");
            if (p != nullptr)
                p->serialize(*this);
            m_w->writeEndElement();
        } else {
            goToStartElementOfTag(tag);
            const bool valid = m_r->attributes().value("valid").toUInt() > 0;
            p = valid ? new ItemClass() : nullptr;
            if (p != nullptr)
                p->serialize(*this);
            goToEndElementOfTag(tag);
        }
    }

    //! serialize item which has only one exact class, and can be present or not
    template <typename ItemClass> void rw(const QString& tag, std::unique_ptr<ItemClass>& up)
    {
        if (m_w) {
            m_w->writeStartElement(tag);
            m_w->writeAttribute("valid", up ? "1" : "0");
            if (up)
                up->serialize(*this);
            m_w->writeEndElement();
        } else {
            goToStartElementOfTag(tag);
            const bool valid = m_r->attributes().value("valid").toUInt() > 0;
            if (valid) {
                up.reset(new ItemClass());
                up->serialize(*this);
            } else
                up.reset();
            goToEndElementOfTag(tag);
        }
    }

    template <typename Catalog> void write(const QString& tag, typename Catalog::CatalogedType*& p)
    {
        ASSERT(m_w);
        m_w->writeStartElement(tag);
        m_w->writeAttribute("type", QString::number(static_cast<uint8_t>(Catalog::type(p))));
        if (p != nullptr)
            p->serialize(*this);
        m_w->writeEndElement();
    }

    template <typename Catalog> void read(const QString& tag, typename Catalog::CatalogedType*& p)
    {
        ASSERT(m_r);
        goToStartElementOfTag(tag);
        const auto type =
            static_cast<typename Catalog::Type>(m_r->attributes().value("type").toUInt());
        p = Catalog::create(type);
        if (p != nullptr)
            p->serialize(*this);
        goToEndElementOfTag(tag);
    }


private:
    void assertCurrentTag(const QString& expectedTag) const;
    void assertCurrentToken(QXmlStreamReader::TokenType token) const;
    void goToEndElementOfTag(const QString& tag);
    void goToStartElementOfTag(const QString& tag);

private:
    QXmlStreamWriter* m_w = nullptr;
    QXmlStreamReader* m_r = nullptr;
};


#endif // BORNAGAIN_GUI_MODEL_SESSION_SERIALIZER_H
