#include "GUI/Model/Group/PropertyItem.h"
#include "GUI/Model/Instrument/BackgroundItems.h"
#include "GUI/Model/Instrument/InstrumentItems.h"
#include "GUI/Model/Instrument/InstrumentModel.h"
#include "GUI/Model/Job/JobModel.h"
#include "GUI/Model/Sample/FormFactorItems.h"
#include "GUI/Model/Sample/LayerItem.h"
#include "GUI/Model/Sample/MultiLayerItem.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Model/Sample/ParticleLayoutItem.h"
#include "GUI/Model/Sample/SampleModel.h"
#include "GUI/Model/Session/Serializer.h"
#include "GUI/Model/Session/SessionModel.h"
#include "Tests/GTestWrapper/google_test.h"
#include <QXmlStreamReader>
#include <QXmlStreamWriter>
#include <memory>

namespace {

QString itemToXML(SessionItem* item)
{
    QString result;
    QXmlStreamWriter writer(&result);
    GUI::Session::XML::writeModel(&writer, item);
    return result;
}

void itemFromXML(QString buffer, SessionItem* item)
{
    QXmlStreamReader reader(buffer);
    GUI::Session::XML::readItems(&reader, item);
}

template <typename ItemType> QByteArray serialize(ItemType& t)
{
    QByteArray a;
    QXmlStreamWriter w(&a);
    w.writeStartElement("start");
    Serializer s(&w);
    t.serialize(s);
    w.writeEndElement();
    return a;
}

template <typename ItemType> void deserialize(ItemType& t, const QByteArray& a)
{
    QXmlStreamReader r(a);
    Serializer sr(&r);
    r.readNextStartElement();
    ASSERT_EQ(r.name().toString(), "start");
    t.serialize(sr);
}

} // namespace

class TestSessionXML : public ::testing::Test {
};

//! Testing to/from xml: simple property item.

TEST_F(TestSessionXML, sessionItem)
{
    QString expected;

    SessionModel source("TestModel");
    source.insertItem<PropertyItem>();

    expected = "<TestModel>"
               "<Item ModelType=\"Property\" Tag=\"rootTag\" DisplayName=\"Property\"/>"
               "</TestModel>";
    EXPECT_EQ(itemToXML(source.rootItem()), expected);

    SessionModel target("TestModel");
    itemFromXML(expected, target.rootItem());

    EXPECT_EQ(target.rowCount(QModelIndex()), 1);
    SessionItem* newItem = target.itemForIndex(target.index(0, 0, QModelIndex()));
    EXPECT_TRUE(newItem->hasModelType<PropertyItem>());
    EXPECT_EQ(newItem->displayName(), "Property");
    EXPECT_FALSE(newItem->value().isValid());
}

//! Testing to/from xml: FullSphereItem

TEST_F(TestSessionXML, InstrumentItem)
{
    // source model, to xml
    InstrumentModel source;
    auto* instrument = source.insertItem<SpecularInstrumentItem>();
    SessionItem* backgroundItem = instrument->backgroundItem();
    QString buffer = itemToXML(source.rootItem());

    // target model, from xml
    InstrumentModel target;
    itemFromXML(buffer, target.rootItem());

    // checking top items in source and target models
    SessionItem* t_item = target.topItem();
    EXPECT_TRUE(t_item->hasModelType<SpecularInstrumentItem>());
    auto* t_sphere = dynamic_cast<SpecularInstrumentItem*>(t_item);
    SessionItem* t_radius = t_sphere->backgroundItem();

    EXPECT_EQ(instrument->parent()->tagFromItem(instrument),
              t_sphere->parent()->tagFromItem(t_sphere));
    EXPECT_EQ(instrument->displayName(), t_sphere->displayName());
    EXPECT_EQ(instrument->modelType(), t_sphere->modelType());
    EXPECT_EQ(instrument->numberOfChildren(), t_sphere->numberOfChildren());
    EXPECT_EQ(instrument->analyzerEfficiency(), t_sphere->analyzerEfficiency());

    EXPECT_EQ(backgroundItem->parent()->tagFromItem(instrument),
              t_radius->parent()->tagFromItem(t_sphere));
    EXPECT_EQ(backgroundItem->displayName(), t_radius->displayName());
    EXPECT_EQ(backgroundItem->modelType(), t_radius->modelType());
    EXPECT_EQ(backgroundItem->numberOfChildren(), t_radius->numberOfChildren());
    EXPECT_EQ(backgroundItem->value().toDouble(), t_radius->value().toDouble());

    // final XML comparison
    EXPECT_EQ(buffer, itemToXML(target.rootItem()));
}

TEST_F(TestSessionXML, emptyMultiLayer)
{
    MultiLayerItem source;
    const auto a1 = serialize(source);

    MultiLayerItem target;
    deserialize(target, a1);

    // final XML comparison
    const auto a2 = serialize(target);
    EXPECT_EQ(a1, a2);
}

TEST_F(TestSessionXML, Layer)
{
    LayerItem source;
    const auto a1 = serialize(source);

    LayerItem target;
    deserialize(target, a1);

    // final XML comparison
    EXPECT_EQ(a1, serialize(target));
}

TEST_F(TestSessionXML, Particle)
{
    ParticleItem source;
    const auto a1 = serialize(source);

    ParticleItem target;
    deserialize(target, a1);

    // final XML comparison
    EXPECT_EQ(a1, serialize(target));
}

TEST_F(TestSessionXML, ParticleWithFF)
{
    ParticleItem source;
    source.setFormFactorType<AnisoPyramidItem>();
    const auto a1 = serialize(source);

    ParticleItem target;
    deserialize(target, a1);

    // final XML comparison
    EXPECT_EQ(a1, serialize(target));
}
